/*
 * Copyright (C) 2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jaguar;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;

import net.sf.cglib.core.DefaultNamingPolicy;
import net.sf.cglib.core.NamingPolicy;
import net.sf.cglib.proxy.Enhancer;

import org.eiichiro.jaguar.inject.Key;
import org.eiichiro.jaguar.interceptor.Dispatcher;
import org.eiichiro.jaguar.lifecycle.Constructed;
import org.eiichiro.jaguar.lifecycle.Event;
import org.eiichiro.jaguar.validation.Constraint;
import org.eiichiro.jaguar.validation.Validator;
import org.eiichiro.jaguar.validation.ViolationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Assembler} assembles a component instance according to the description.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Assembler<T> {

	private static final NamingPolicy NAMING_POLICY = new DefaultNamingPolicy() {
		
		@Override
		protected String getTag() {
			return "ByJaguar";
		}
		
	};
	
	private Logger logger = LoggerFactory.getLogger(getClass());
	
	private final Container container;
	
	private final Description<T> description;
	
	private T instance;
	
	/**
	 * Constructs a new {@code Assembler} instance with the specified container 
	 * and description.
	 * 
	 * @param container The {@link Container} dependent components are required.
	 * @param description The component description.
	 */
	public Assembler(Container container, Description<T> description) {
		this.container = container;
		this.description = description;
	}
	
	/**
	 * Constructs a new {@code Assembler} instance with the specified container, 
	 * description and component instance.
	 * 
	 * @param container The {@link Container} dependent components are required.
	 * @param description The component description.
	 * @param instance The component instance to be assembled.
	 */
	public Assembler(Container container, Description<T> description, T instance) {
		this.container = container;
		this.description = description;
		this.instance = instance;
	}
	
	/**
	 * Assembles a component instance according to the description.
	 * {@code Assembler} assembles a component instance by the following 
	 * instructions: 
	 * <ol>
	 * <li>Instance construction. If the component instance has not been 
	 * specified, assembler constructs the instance first. When a joinpoint 
	 * method is declared on the component class, assembler generates the 
	 * subclass dynamically with <a href="http://cglib.sourceforge.net/">cglib</a> 
	 * and instantiates it as the component instance attaching {@link Dispatcher} 
	 * as the method invocation handler to intercept the method invocations. 
	 * Otherwise, assembler constructs the instance invoking {@code Class#newInstance()} 
	 * method on the class {@code Description#type()} returns. After the instance 
	 * construction, assembler fires the {@link Constructed} event on the 
	 * instance. If any exception has occurred, assembler returns <code>null</code> 
	 * immediately.</li>
	 * <li>Dependency injection. Assembler gets the dependent component from the 
	 * {@code Container} corresponding to the {@link Key} constructed from the 
	 * injective field and populates it to the assembling instance. If any 
	 * exception has occurred, assembler skips the failed population.
	 * </li>
	 * <li>Validation. After the dependencies are resolved, assembler validates 
	 * every constrained field and the component instance by invoking 
	 * {@code Validator#valid(Annotation, Object)}. If the method returns 
	 * <code>false</code>, assembler throws {@link ViolationException}. If all 
	 * constrained fields are validated, the assembled instance is returned.</li>
	 * </ol>
	 * 
	 * @return Component instance assembled.
	 */
	@SuppressWarnings("unchecked")
	public T assemble() {
		if (instance == null) {
			try {
				if (description.joinpoints().isEmpty()) {
					instance = description.type().newInstance();
				} else {
					Enhancer enhancer = new Enhancer();
					enhancer.setSuperclass(description.type());
					enhancer.setCallback(new Dispatcher(container, description));
					enhancer.setNamingPolicy(NAMING_POLICY);
					instance = (T) enhancer.create();
				}
				
				container.descriptions().put(instance, description);
				new Event<Constructed>() {}.on(instance).fire();
			} catch (Exception e) {
				logger.warn("Component [" + description + "] cannot be instantiated", e);
				return null;
			}
		}
		
		for (Field field : description.injects()) {
			if (!field.isAccessible()) {
				field.setAccessible(true);
			}
			
			try {
				field.set(instance, container.component(new Key<Object>(field)));
			} catch (Exception e) {
				logger.warn("Dependent component cannot be injected into [" + field + "]", e);
			}
		}
		
		for (Field field : description.validates()) {
			for (Annotation annotation : field.getAnnotations()) {
				Class<? extends Annotation> annotationType = annotation.annotationType();
				
				if (annotationType.isAnnotationPresent(Constraint.class)) {
					Constraint constraint = annotationType.getAnnotation(Constraint.class);
					Class<? extends Validator<Annotation>> validator = (Class<? extends Validator<Annotation>>) constraint.value();
					Validator<Annotation> v = container.component(validator);
					
					if (v == null) {
						container.install(validator);
						v = container.component(validator);
					}
					
					if (!field.isAccessible()) {
						field.setAccessible(true);
					}
					
					Object value = null;
					
					try {
						value = field.get(instance);
					} catch (Exception e) {
						logger.warn("Component field [" + field + "] cannot be validated", e);
					}
					
					boolean valid = false;
					
					try {
						valid = v.valid(annotation, value);
					} catch (ClassCastException e) {
						throw new ConfigurationException(
								"The other way around about Constraint and Validator: Validator ["
										+ validator + "] must validate ["
										+ annotationType + "]", e);
					}
					
					if (valid) {
						logger.debug("Field [" + field + "]'s value [" + value + "] is valid for ["
								+ annotation + "] constraint");
					} else {
						logger.error("Field [" + field + "]'s value [" + value + "] is invalid for ["
								+ annotation + "] constraint");
						throw new ViolationException(field, annotation, value);
					}
				}
			}
		}
		
		for (Annotation annotation : description.constraints()) {
			Class<? extends Annotation> annotationType = annotation.annotationType();
			Constraint constraint = annotationType.getAnnotation(Constraint.class);
			Class<? extends Validator<Annotation>> validator = (Class<? extends Validator<Annotation>>) constraint.value();
			Validator<Annotation> v = container.component(validator);
			
			if (v == null) {
				container.install(validator);
				v = container.component(validator);
			}
			
			boolean valid = false;
			
			try {
				valid = v.valid(annotation, instance);
			} catch (ClassCastException e) {
				throw new ConfigurationException(
						"The other way around about Constraint and Validator: Validator ["
								+ validator + "] must validate ["
								+ annotationType + "]", e);
			}
			
			if (valid) {
				logger.debug("Class [" + description.type() + "]'s value [" 
						+ instance + "] is valid for [" + annotation + "] constraint");
			} else {
				logger.error("Class [" + description.type() + "]'s value [" 
						+ instance + "] is invalid for [" + annotation + "] constraint");
				throw new ViolationException(description.type(), annotation, instance);
			}
		}
		
		logger.debug("Component [" + description + "] has been assembled");
		return instance;
	}
	
}
