/*
 * Copyright (C) 2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jaguar.inject;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.HashSet;
import java.util.Set;

import org.eiichiro.jaguar.Assembler;

import com.google.common.base.Preconditions;

/**
 * Component type and binding annotations to lookup the component from the 
 * container.
 * {@code Key} is constructed by the {@link Assembler} internally from the 
 * {@code @Inject}-annotated field that the dependent component to be injected.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Key<T> {

	private final Class<T> type;
	
	private final Set<Annotation> bindings;
	
	/**
	 * Constructs a new {@code Key} instance from the specified injective 
	 * {@code Field}.
	 * 
	 * @param field The {@code Field} the dependent component is injected.
	 */
	@SuppressWarnings("unchecked")
	public Key(Field field) {
		Preconditions.checkArgument(field != null, 
				"Parameter 'field' must not be [" + field + "]");
		type = (Class<T>) field.getType();
		Set<Annotation> bindings = new HashSet<Annotation>();
		
		for (Annotation annotation : field.getAnnotations()) {
			if (annotation.annotationType().getAnnotation(Binding.class) != null) {
				bindings.add(annotation);
			}
		}
		
		this.bindings = bindings;
	}
	
	/**
	 * Returns the class to qualify the component to be injected.
	 * 
	 * @return The class to qualify the component to be injected.
	 */
	public Class<T> type() {
		return type;
	}
	
	/**
	 * Returns the bindings to qualify the component to be injected.
	 * 
	 * @return The bindings to qualify the component to be injected.
	 */
	public Set<Annotation> bindings() {
		return bindings;
	}
	
	/** String representation. */
	@Override
	public String toString() {
		return "type: " + type + ", bindings: " + bindings;
	}
	
}
